package server

import (
	pb "asapo_common/generated_proto"
	log "asapo_common/logger"
	"context"
	"errors"
	influxdb2 "github.com/influxdata/influxdb-client-go/v2"
	"google.golang.org/grpc"
	"net"
	"strconv"
	"time"
)

func GetSwitshPeriod(retentionTime string, groupingTime string) (int64, int64, error) {
	// Return time, at which Querry server switch from full to reduce DB
	retentionTimeInt, err := time.ParseDuration(retentionTime)
	if err != nil {
		return 0, 0, err
	}
	groupingTimeInt, err := time.ParseDuration(groupingTime)
	if err != nil {
		return 0, 0, err
	}

	switchPeriod := retentionTimeInt.Seconds() - groupingTimeInt.Seconds()
	if switchPeriod < 0 {
		return 0, 0, errors.New("retention time is less then grouping time")
	}
	return int64(switchPeriod), int64(groupingTimeInt.Seconds()), nil
}

func Start(settings Settings) {
	lis, err := net.Listen("tcp", ":"+strconv.Itoa(int(settings.ServerPort)))
	if err != nil {
		log.Fatal("failed to listen: ", err)
	}

	influxClient := influxdb2.NewClient(settings.InfluxDbUrl, "" /*settings.InfluxDb2AuthToken*/)
	// influxClient.BucketsAPI().CreateBucketWithName(context.Background(), domain.Organization{}, "bucketName", domain.RetentionRules{})

	queryServer := QueryServer{}
	ingestServer := IngestServer{}

	dbQueryApi := influxClient.QueryAPI("" /*settings.InfluxDb2Org*/)
	queryServer.dbQueryApi = dbQueryApi
	queryServer.settings = settings

	switchPeriod, groupedInterval, err := GetSwitshPeriod(settings.RetentionPolicyTime, settings.GroupingTime)
	if err != nil {
		log.Fatal("failed to get switch period", err)
		return
	}
	queryServer.switchPeriod = switchPeriod
	queryServer.groupedInterval = groupedInterval

	dbWriterApi := influxClient.WriteAPI("" /*settings.InfluxDb2Org*/, settings.InfluxDbDatabase)
	ingestServer.dbWriterApi = dbWriterApi
	ingestServer.settings = settings

	grpcServer := grpc.NewServer()
	pb.RegisterAsapoMonitoringQueryServiceServer(grpcServer, &queryServer)
	pb.RegisterAsapoMonitoringIngestServiceServer(grpcServer, &ingestServer)
	log.Info("server listening at ", lis.Addr())

	// Check connection
	_, err = queryServer.GetMetadata(context.TODO(), &pb.Empty{})
	if err != nil {
		log.Fatal("dummy database request failed: ", err)
	}

	// Start net server and wait
	if err = grpcServer.Serve(lis); err != nil {
		log.Fatal("failed to serve: ", err)
	}
}
